import asyncio
import logging
import re
from typing import Dict, List

from container_app_conf.formatter.toml import TomlFormatter
from requests import HTTPError
from telegram import Update, InlineKeyboardMarkup, InlineKeyboardButton, ReplyKeyboardRemove
from telegram.ext import CommandHandler, filters, MessageHandler, CallbackQueryHandler, \
    ApplicationBuilder, ContextTypes
from telegram_click.argument import Argument, Flag
from telegram_click.decorator import command

from keel_telegram_bot import util
from keel_telegram_bot.api_client import KeelApiClient
from keel_telegram_bot.bot.permissions import CONFIG_ADMINS, CONFIGURED_CHAT_ID
from keel_telegram_bot.bot.reply_keyboard_handler import ReplyKeyboardHandler
from keel_telegram_bot.config import Config
from keel_telegram_bot.stats import *
from keel_telegram_bot.util import send_message, approval_to_str

LOGGER = logging.getLogger(__name__)


class KeelTelegramBot:
    """
    The main entry class of the keel telegram bot
    """

    def __init__(self, config: Config, api_client: KeelApiClient):
        """
        Creates an instance.
        :param config: configuration object
        """
        self._config = config
        self._api_client = api_client
        self._message_map = {}

        self._response_handler = ReplyKeyboardHandler()

        self._app = ApplicationBuilder().token(self._config.TELEGRAM_BOT_TOKEN.value).build()

        handler_groups = {
            0: [CallbackQueryHandler(callback=self._inline_keyboard_click_callback)],
            1: [
                CommandHandler(COMMAND_START,
                               filters=(~ filters.REPLY) & (~ filters.FORWARDED),
                               callback=self._start_callback),
                CommandHandler(COMMAND_LIST_APPROVALS,
                               filters=(~ filters.REPLY) & (~ filters.FORWARDED),
                               callback=self._list_approvals_callback),
                CommandHandler(COMMAND_APPROVE,
                               filters=(~ filters.REPLY) & (~ filters.FORWARDED),
                               callback=self._approve_callback),
                CommandHandler(COMMAND_REJECT,
                               filters=(~ filters.REPLY) & (~ filters.FORWARDED),
                               callback=self._reject_callback),
                CommandHandler(COMMAND_DELETE,
                               filters=(~ filters.REPLY) & (~ filters.FORWARDED),
                               callback=self._delete_callback),
                CommandHandler(COMMAND_CHATID,
                               filters=(~ filters.REPLY) & (~ filters.FORWARDED),
                               callback=self._chatid_callback),

                CommandHandler(COMMAND_HELP,
                               filters=(~ filters.REPLY) & (~ filters.FORWARDED),
                               callback=self._help_callback),
                CommandHandler(COMMAND_CONFIG,
                               filters=(~ filters.REPLY) & (~ filters.FORWARDED),
                               callback=self._config_callback),
                CommandHandler(COMMAND_VERSION,
                               filters=(~ filters.REPLY) & (~ filters.FORWARDED),
                               callback=self._version_callback),
                CommandHandler(CANCEL_KEYBOARD_COMMAND[1:],
                               filters=(~ filters.REPLY) & (~ filters.FORWARDED),
                               callback=self._response_handler.cancel_keyboard_callback),
                # unknown command handler
                MessageHandler(
                    filters=filters.COMMAND & (~ filters.FORWARDED),
                    callback=self._unknown_command_callback),
                MessageHandler(
                    filters=(~ filters.FORWARDED),
                    callback=self._any_message_callback),
            ]
        }

        for group, handlers in handler_groups.items():
            for handler in handlers:
                self._app.add_handler(handler, group=group)

    @property
    def bot(self):
        return self._app.bot

    async def start(self):
        """
        Starts up the bot.
        """
        await self._app.initialize()
        LOGGER.debug(f"Using bot id '{self._app.bot.id}' ({self._app.bot.name})")
        await self._app.start()
        await self._app.updater.start_polling()

    def stop(self):
        """
        Shuts down the bot.
        """
        loop = asyncio.get_event_loop()
        loop.run_until_complete(self._app.shutdown())

    @COMMAND_TIME_START.time()
    async def _start_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
        """
        Welcomes a new user with a greeting message
        :param update: the chat update object
        :param context: telegram context
        """
        bot = context.bot
        chat_id = update.effective_chat.id
        user_first_name = update.effective_user.first_name

        if not CONFIG_ADMINS.evaluate(update, context):
            await send_message(bot, chat_id, "Sorry, you do not have permissions to use this bot.")
            return

        await send_message(bot, chat_id,
                           f"Welcome {user_first_name},\nthis is your keel-telegram-bot instance, ready to go!")

    @COMMAND_TIME_LIST_APPROVALS.time()
    @command(name=COMMAND_LIST_APPROVALS,
             description="List pending approvals",
             arguments=[
                 Flag(name=["archived", "h"], description="Include archived items"),
                 Flag(name=["approved", "a"], description="Include approved items"),
                 Flag(name=["rejected", "r"], description="Include rejected items"),
             ],
             permissions=CONFIGURED_CHAT_ID & CONFIG_ADMINS)
    async def _list_approvals_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE,
                                       archived: bool, approved: bool, rejected: bool) -> None:
        """
        List pending approvals
        """
        bot = context.bot
        message = update.effective_message
        chat_id = update.effective_chat.id

        items = self._api_client.get_approvals()
        items = list(filter(lambda x: not self._is_filtered_for(chat_id, x["identifier"]), items))

        rejected_items = list(filter(lambda x: x[KEY_REJECTED], items))
        archived_items = list(filter(lambda x: x[KEY_ARCHIVED], items))
        pending_items = list(filter(
            lambda x: x not in archived_items and x not in rejected_items
                      and x[KEY_VOTES_RECEIVED] < x[KEY_VOTES_REQUIRED], items))
        approved_items = list(
            filter(lambda x: x not in rejected_items and x not in archived_items and x not in pending_items, items))

        lines = []
        if archived:
            lines.append("\n".join([
                f"<b>=== Archived ({len(archived_items)}) ===</b>",
                "",
                "\n\n".join(list(map(lambda x: "> " + approval_to_str(x), archived_items)))
            ]).strip())

        if approved:
            lines.append("\n".join([
                f"<b>=== Approved ({len(approved_items)}) ===</b>",
                "",
                "\n\n".join(list(map(lambda x: "> " + approval_to_str(x), approved_items))),
            ]).strip())

        if rejected:
            lines.append("\n".join([
                f"<b>=== Rejected ({len(rejected_items)}) ===</b>",
                "",
                "\n\n".join(list(map(lambda x: "> " + approval_to_str(x), rejected_items))),
            ]).strip())

        lines.append("\n".join([
            f"<b>=== Pending ({len(pending_items)}) ===</b>",
            "",
            "\n\n".join(list(map(lambda x: "> " + approval_to_str(x), pending_items))),
        ]))

        text = "\n\n".join(lines).strip()
        await send_message(bot, chat_id, text, reply_to=message.message_id, parse_mode="HTML")

    @COMMAND_TIME_APPROVE.time()
    @command(name=COMMAND_APPROVE,
             description="Approve a pending item",
             arguments=[
                 Argument(name=["identifier", "i"], description="Approval identifier or id",
                          example="default/myimage:1.5.5"),
                 Argument(name=["voter", "v"], description="Name of voter", example="john", optional=True),
             ],
             permissions=CONFIGURED_CHAT_ID & CONFIG_ADMINS)
    async def _approve_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE,
                                identifier: str, voter: str or None) -> None:
        """
        Approve a pending item
        """
        chat_id = update.effective_chat.id

        if voter is None:
            voter = update.effective_user.full_name

        async def execute(update: Update, context: ContextTypes.DEFAULT_TYPE, item: dict, data: dict):
            bot = context.bot
            message = update.effective_message
            chat_id = update.effective_chat.id

            self._api_client.approve(item["id"], item["identifier"], voter)
            text = f"Approved {item['identifier']}"
            await send_message(bot, chat_id, text, reply_to=message.message_id,
                               menu=ReplyKeyboardRemove(selective=True))

        items = self._api_client.get_approvals(rejected=False, archived=False)
        items = list(filter(lambda x: not self._is_filtered_for(chat_id, x["identifier"]), items))

        # compare to the "id" first
        exact_matches = list(filter(lambda x: x["id"] == identifier, items))
        if len(exact_matches) > 0:
            await execute(update, context, exact_matches[0], {})
            return

        # then fuzzy match to "identifier"
        await self._response_handler.await_user_selection(
            update, context, identifier, choices=items, key=lambda x: x["identifier"],
            callback=execute,
        )

    @COMMAND_TIME_REJECT.time()
    @command(name=COMMAND_REJECT,
             description="Reject a pending item",
             arguments=[
                 Argument(name=["identifier", "i"], description="Approval identifier or id",
                          example="default/myimage:1.5.5"),
                 Argument(name=["voter", "v"], description="Name of voter", example="john", optional=True),
             ],
             permissions=CONFIGURED_CHAT_ID & CONFIG_ADMINS)
    async def _reject_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE,
                               identifier: str, voter: str or None) -> None:
        """
        Reject a pending item
        """
        chat_id = update.effective_chat.id
        if voter is None:
            voter = update.effective_user.full_name
        if not voter:
            voter = update.effective_user.name

        async def execute(update: Update, context: ContextTypes.DEFAULT_TYPE, item: dict, data: dict):
            bot = context.bot
            message = update.effective_message
            chat_id = update.effective_chat.id

            self._api_client.reject(item["id"], item["identifier"], voter)
            text = f"Rejected {item['identifier']}"
            await send_message(bot, chat_id, text, reply_to=message.message_id,
                               menu=ReplyKeyboardRemove(selective=True))

        items = self._api_client.get_approvals(rejected=False, archived=False)
        items = list(filter(lambda x: not self._is_filtered_for(chat_id, x["identifier"]), items))

        # compare to the "id" first
        exact_matches = list(filter(lambda x: x["id"] == identifier, items))
        if len(exact_matches) > 0:
            await execute(update, context, exact_matches[0], {})
            return

        # then fuzzy match to "identifier"
        await self._response_handler.await_user_selection(
            update, context, identifier, choices=items, key=lambda x: x["identifier"],
            callback=execute,
        )

    @COMMAND_TIME_DELETE.time()
    @command(name=COMMAND_DELETE,
             description="Delete an approval item",
             arguments=[
                 Argument(name=["identifier", "i"], description="Approval identifier or id",
                          example="default/myimage:1.5.5"),
                 Argument(name=["voter", "v"], description="Name of voter", example="john", optional=True),
             ],
             permissions=CONFIGURED_CHAT_ID & CONFIG_ADMINS)
    async def _delete_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE,
                               identifier: str, voter: str or None) -> None:
        """
        Delete an archived item
        """
        chat_id = update.effective_chat.id
        if voter is None:
            voter = update.effective_user.full_name

        async def execute(update: Update, context: ContextTypes.DEFAULT_TYPE, item: dict, data: dict):
            bot = context.bot
            message = update.effective_message
            chat_id = update.effective_chat.id

            self._api_client.delete(item["id"], item["identifier"], voter)
            text = f"Deleted {item['identifier']}"
            await send_message(bot, chat_id, text, reply_to=message.message_id,
                               menu=ReplyKeyboardRemove(selective=True))

        items = self._api_client.get_approvals()
        items = list(filter(lambda x: not self._is_filtered_for(chat_id, x["identifier"]), items))

        # compare to the "id" first
        exact_matches = list(filter(lambda x: x["id"] == identifier, items))
        if len(exact_matches) > 0:
            await execute(update, context, exact_matches[0], {})
            return

        # then fuzzy match to "identifier"
        await self._response_handler.await_user_selection(
            update, context, identifier, choices=items, key=lambda x: x["identifier"],
            callback=execute,
        )

    async def on_notification(self, data: dict):
        """
        Handles incoming notifications (via Webhook)
        :param data: notification data
        """
        KEEL_NOTIFICATION_COUNTER.inc()

        identifier = data.get("identifier", None)
        title = data.get("name", None)
        type = data.get("type", None)
        level = data.get("level", None)  # success/failure
        message = data.get("message", None)

        text = "\n".join([
            f"<b>{title}: {level}</b>",
            f"{identifier}",
            f"{type}",
            f"{message}",
        ])

        for chat_id in self._config.TELEGRAM_CHAT_IDS.value:

            if self._is_filtered_for(chat_id, identifier):
                continue

            await send_message(
                self.bot, chat_id,
                text, parse_mode="HTML",
                menu=None
            )

    async def on_new_pending_approval(self, item: dict):
        """
        Handles new pending approvals by sending a message
        including an inline keyboard to all configured chat ids
        :param item: new pending approval
        """
        identifier = item["identifier"]
        text = approval_to_str(item)
        menu = self.create_approval_notification_menu(item)

        for chat_id in self._config.TELEGRAM_CHAT_IDS.value:

            if self._is_filtered_for(chat_id, identifier):
                LOGGER.debug(f"Skipping new pending approval for chat '{chat_id}' due to filters")
                continue

            LOGGER.debug(f"Sending pending approval message to '{chat_id}'")
            try:
                response = await send_message(
                    self.bot, chat_id,
                    text, parse_mode="HTML",
                    menu=menu
                )
                self._register_message(response.chat_id, response.message_id, item["id"], item["identifier"])
            except Exception as ex:
                LOGGER.exception(ex)

    @command(
        name=COMMAND_CONFIG,
        description="Print bot config.",
        permissions=CONFIGURED_CHAT_ID & CONFIG_ADMINS,
    )
    async def _config_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        bot = context.bot
        message = update.effective_message
        chat_id = update.effective_chat.id
        text = self._config.print(TomlFormatter())
        await send_message(bot, chat_id, text, reply_to=message.message_id)

    @command(
        name=COMMAND_HELP,
        description="List commands supported by this bot.",
        permissions=CONFIG_ADMINS,
    )
    async def _help_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        bot = context.bot
        message = update.effective_message
        chat_id = update.effective_chat.id

        from telegram_click import generate_command_list
        text = await generate_command_list(update, context)
        await send_message(bot, chat_id, text,
                           parse_mode="MARKDOWN",
                           reply_to=message.message_id)

    @command(
        name=COMMAND_CHATID,
        description="Print chat ID.",
        permissions=CONFIG_ADMINS,
    )
    async def _chatid_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        bot = context.bot
        message = update.effective_message
        chat_id = update.effective_chat.id

        text = update.message.chat_id
        await send_message(bot, chat_id, str(text), reply_to=message.message_id)

    @command(
        name=COMMAND_VERSION,
        description="Print bot version.",
        permissions=CONFIG_ADMINS,
    )
    async def _version_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        bot = context.bot
        message = update.effective_message
        chat_id = update.effective_chat.id

        from keel_telegram_bot import __version__
        text = __version__
        await send_message(bot, chat_id, text, reply_to=message.message_id)

    async def _unknown_command_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
        """
        Handles unknown commands send by a user
        :param update: the chat update object
        :param context: telegram context
        """
        message = update.effective_message
        username = "N/A"
        if update.effective_user is not None:
            username = update.effective_user.username

        user_is_admin = username in self._config.TELEGRAM_ADMIN_USERNAMES.value
        if user_is_admin:
            await self._help_callback(update, context)
            return

    async def _any_message_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE) -> None:
        """
        Used to respond to response keyboard entry selections
        :param update: the chat update object
        :param context: telegram context
        """
        await self._response_handler.on_message(update, context)

    async def _inline_keyboard_click_callback(self, update: Update, context: ContextTypes.DEFAULT_TYPE):
        """
        Handles inline keyboard button click callbacks
        :param update:
        :param context:
        """
        bot = context.bot
        from_user = update.callback_query.from_user

        message_text = update.effective_message.text
        query = update.callback_query
        query_id = query.id
        data = query.data

        if data == BUTTON_DATA_NOTHING:
            return

        try:
            matches = re.search(r"^Id: (.*)", message_text, flags=re.MULTILINE)
            approval_id = matches.group(1)
            matches = re.search(r"^Identifier: (.*)", message_text, flags=re.MULTILINE)
            approval_identifier = matches.group(1)

            if data == BUTTON_DATA_APPROVE:
                self._api_client.approve(approval_id, approval_identifier, from_user.full_name)
                answer_text = f"Approved '{approval_identifier}'"
                KEEL_APPROVAL_ACTION_COUNTER.labels(action="approve", identifier=approval_identifier).inc()
            elif data == BUTTON_DATA_REJECT:
                self._api_client.reject(approval_id, approval_identifier, from_user.full_name)
                answer_text = f"Rejected '{approval_identifier}'"
                KEEL_APPROVAL_ACTION_COUNTER.labels(action="reject", identifier=approval_identifier).inc()
            else:
                await bot.answer_callback_query(query_id, text="Unknown button")
                return

            await context.bot.answer_callback_query(query_id, text=answer_text)
            await self.update_messages()
        except HTTPError as e:
            LOGGER.error(e)
            await bot.answer_callback_query(query_id, text=f"{e.response.content.decode('utf-8')}")
        except Exception as e:
            LOGGER.error(e)
            await bot.answer_callback_query(query_id, text=f"Unknwon error")

    @staticmethod
    def _build_inline_keyboard(items: Dict[str, str]) -> InlineKeyboardMarkup:
        """
        Builds an inline button menu
        :param items: dictionary of "button text" -> "callback data" items
        :return: reply markup
        """
        keyboard = list(map(lambda x: InlineKeyboardButton(x[0], callback_data=x[1]), items.items()))
        return InlineKeyboardMarkup.from_column(keyboard)

    def create_approval_notification_menu(self, item: dict) -> InlineKeyboardMarkup:
        keyboard_items = {}
        if item["archived"]:
            keyboard_items["Approved"] = BUTTON_DATA_NOTHING
        elif item["rejected"]:
            keyboard_items["Rejected"] = BUTTON_DATA_NOTHING
        else:
            if item["votesRequired"] > item["votesReceived"]:
                keyboard_items["Approve"] = BUTTON_DATA_APPROVE
                keyboard_items["Reject"] = BUTTON_DATA_REJECT

        return self._build_inline_keyboard(keyboard_items)

    def _register_message(self, chat_id: int, message_id: int, approval_id: str, approval_identifier: str):
        """
        Registers a telegram message, that corresponds with an approval notification.
        This is used to update this message. This is possible for approx. 48 hours, after
        which telegram prohibits modifications of the original message.
        :param chat_id: chat id
        :param message_id: message id
        :param approval_id: approval id
        :param approval_identifier: approval identifier
        """
        key = f"{approval_id}_{approval_identifier}"
        self._message_map.setdefault(key, {}).setdefault(chat_id, set()).add(message_id)

    async def update_messages(self):
        """
        Fetch approvals and update existing approval messages accordingly
        """
        approvals = self._api_client.get_approvals()

        for approval in approvals:
            approval_id = approval["id"]
            approval_identifier = approval["identifier"]
            key = f"{approval_id}_{approval_identifier}"

            chats = self._message_map.get(key, {})
            failed_messages = set()
            for chat_id, message_ids in chats.items():

                if self._is_filtered_for(chat_id, approval_identifier):
                    continue

                for message_id in message_ids:
                    try:
                        approval_str = approval_to_str(approval)
                        menu = self.create_approval_notification_menu(approval)
                        await self.bot.edit_message_text(
                            approval_str,
                            chat_id=chat_id,
                            message_id=message_id,
                            parse_mode="HTML",
                            reply_markup=menu
                        )
                    except Exception as ex:
                        failed_messages.add(message_id)
                        LOGGER.exception(ex)

                for failure in failed_messages:
                    message_ids.remove(failure)

    def _is_filtered_for(self, chat_id: str or int, identifier: str) -> bool:
        chat_ids: List[str] = self._config.TELEGRAM_CHAT_IDS.value
        filter_config: List[Dict] = self._config.TELEGRAM_FILTERS.value

        chat_unknown = str(chat_id) not in chat_ids
        filter_doesnt_match = util._is_filtered_for(filter_config, str(chat_id), identifier)

        result = chat_unknown or filter_doesnt_match
        LOGGER.debug(
            f"Filtered identifier '{identifier}' for chat {chat_id} because: chat_unknown: {chat_unknown}, filter_doesnt_match: {filter_doesnt_match}, result: {result}")

        return result



REQUESTS_TIMEOUT = (5, 5)
TELEGRAM_CAPTION_LENGTH_LIMIT = 200

# Commands
COMMAND_START = "start"
COMMAND_LIST_APPROVALS = ["list", "l"]
COMMAND_APPROVE = ["approve", "a"]
COMMAND_REJECT = ["reject", "r"]
COMMAND_DELETE = ["delete", "d"]
COMMAND_CHATID = "chatid"

COMMAND_STATS = 'stats'

COMMAND_HELP = ['help', 'h']
COMMAND_VERSION = ['version', 'v']
COMMAND_CONFIG = ['config', 'c']

CANCEL_KEYBOARD_COMMAND = "/cancel_keyboard"

# inline keyboard button data
BUTTON_DATA_NOTHING = "_"
BUTTON_DATA_APPROVE = "a"
BUTTON_DATA_REJECT = "r"

# webserver

ENDPOINT_WEBHOOK = "/"

# api keys
KEY_ARCHIVED = "archived"
KEY_REJECTED = "rejected"
KEY_VOTES_REQUIRED = "votesRequired"
KEY_VOTES_RECEIVED = "votesReceived"



import logging
from typing import List

import requests
from requests.auth import HTTPBasicAuth

from keel_telegram_bot.const import REQUESTS_TIMEOUT

LOGGER = logging.getLogger(__name__)

GET = "get"
POST = "post"


class KeelApiClient:
    """
    Keel REST api client
    """

    def __init__(self, host: str, port: int, ssl: bool, user: str, password: str):
        self._host = host
        self._port = port
        self._ssl = ssl
        self._auth = HTTPBasicAuth(user, password)

        self._base_url = f"{'https' if ssl else 'http'}://{host}:{port}"

    def get_approvals(self, rejected: bool = None, archived: bool = None) -> List[dict]:
        """
        :param rejected: True for rejected, False for approved, None for all
        :param archived: True for archived, False for not archived, None for all
        :return: a list of all approvals matching criteria
        """
        response = self._do_request(GET, self._base_url + "/v1/approvals")

        if rejected is not None:
            response = list(filter(lambda x: x["rejected"] == rejected, response))
        if archived is not None:
            response = list(filter(lambda x: x["archived"] == archived, response))

        return response

    def approve(self, id: str, identifier: str, voter: str) -> None:
        """
        Approve a pending approval
        :param id: item id
        :param identifier: identifier for the approval request, something like "default/myimage:1.5.5"
        :param voter: name of the voter
        """
        self._do_request(POST, self._base_url + "/v1/approvals", json={
            "id": id,
            "identifier": identifier,
            "action": "approve",
            "voter": voter,
        })

    def reject(self, id: str, identifier: str, voter: str) -> None:
        """
        Reject a pending approval
        :param id: item id
        :param identifier: identifier for the approval request, something like "default/myimage:1.5.5"
        :param voter: name of the voter
        """
        self._do_request(POST, self._base_url + "/v1/approvals", json={
            "id": id,
            "identifier": identifier,
            "action": "reject",
            "voter": voter,
        })

    def delete(self, id: str, identifier: str, voter: str) -> None:
        """
        Delete a pending approval
        :param id: item id
        :param identifier: identifier for the approval request, something like "default/myimage:1.5.5"
        :param voter: name of the voter
        """
        self._do_request(POST, self._base_url + "/v1/approvals", json={
            "id": id,
            "identifier": identifier,
            "action": "delete",
            "voter": voter,
        })

    def _do_request(self, method: str = GET, url: str = "/", params: dict = None,
                    json: dict = None) -> list or dict or None:
        """
        Executes a http request based on the given parameters

        :param method: the method to use (GET, POST)
        :param url: the url to use
        :param params: query parameters that will be appended to the url
        :param json: request body
        :return: the response parsed as a json
        """
        headers = []
        url = self._create_request_url(url, params)

        if method is GET:
            method = requests.get
        elif method is POST:
            method = requests.post
        else:
            raise ValueError("Unsupported method: {}".format(method))

        response = method(url, headers=headers, auth=self._auth, json=json, timeout=REQUESTS_TIMEOUT)

        response.raise_for_status()
        # some responses do not return data so we just ignore the body in that case
        if len(response.content) > 0 and response.content != b"null":
            return response.json()

    @staticmethod
    def _create_request_url(url: str, params: dict = None):
        """
        Adds query params to the given url

        :param url: the url to extend
        :param params: query params as a keyed dictionary
        :return: the url including the given query params
        """
        if params:
            first_param = True
            for k, v in sorted(params.items(), key=lambda entry: entry[0]):
                if not v:
                    # skip None values
                    continue

                if first_param:
                    url += '?'
                    first_param = False
                else:
                    url += '&'

                url += "%s=%s" % (k, v)

        return url



import unittest

from keel_telegram_bot.config import Config


class TestBase(unittest.TestCase):
    from container_app_conf.source.yaml_source import YamlSource

    # load config from test folder
    config = Config(
        singleton=True,
        data_sources=[
            YamlSource("keel-telegram-bot.yaml", "./tests/")
        ]
    )



import functools
import logging
import operator
import re
from datetime import datetime, timezone, timedelta
from typing import List, Any, Tuple, Dict

import iso8601
from telegram import Bot, Message
from telegram._utils.types import ReplyMarkup

from keel_telegram_bot.config import Config

LOGGER = logging.getLogger(__name__)

CONFIG = Config()


def _is_filtered_for(filters: List[Dict], chat_id: str, identifier: str) -> bool:
    for config in filters:
        filter_chat_id = config["chat_id"]
        identifier_regex = config["identifier"]

        if str(filter_chat_id) == str(chat_id):
            identifier_pattern = re.compile(identifier_regex)
            result = identifier_pattern.search(identifier)
            if result is None:
                return True

    return False


def flatten(data: List[List[Any]]) -> List[Any]:
    """
    Flattens a list of lists
    :param data: the data to flatten
    :return: flattened list
    """
    return functools.reduce(operator.iconcat, data, [])


def format_for_single_line_log(text: str) -> str:
    """
    Formats a text for log
    :param text:
    :return:
    """
    text = "" if text is None else text
    return " ".join(text.split())


async def send_message(bot: Bot, chat_id: str, message: str, parse_mode: str = None, reply_to: int = None,
                       menu: ReplyMarkup = None) -> Message:
    """
    Sends a text message to the given chat
    :param bot: the bot
    :param chat_id: the chat product_id to send the message to
    :param message: the message to chat (may contain emoji aliases)
    :param parse_mode: specify whether to parse the text as markdown or HTML
    :param reply_to: the message product_id to reply to
    :param menu: inline keyboard menu markup
    """
    from emoji import emojize

    emojized_text = emojize(message)
    return await bot.send_message(
        chat_id=chat_id, parse_mode=parse_mode, text=emojized_text,
        reply_to_message_id=reply_to,
        reply_markup=menu
    )


def fuzzy_match(term: str, choices: List[Any], limit: int = None, key=lambda x: x, ignorecase: bool = True) -> List[
    Tuple[Any, int]]:
    """
    Does a fuzzy search on the given choices
    :param term: the search term
    :param choices: list of possible choices
    :param key: function to turn a choice item into a string
    :param limit: Optional maximum for the number of elements returned
    :return: List of (choice, ratio) tuples, sorted by descending ratio
    """
    # map choices to key
    if ignorecase:
        term = term.casefold()
    choices = filter(lambda x: key(x) is not None, choices)
    key_map = dict(map(lambda x: (key(x).casefold() if ignorecase else key(x), x), choices))

    from fuzzywuzzy import process
    from fuzzywuzzy import fuzz
    matches = process.extract(term, key_map.keys(), limit=limit, scorer=fuzz.UWRatio)

    # map results back to original choices
    result = list(map(lambda x: (key_map[x[0]], x[1]), matches))

    return result


def filter_new_by_key(a: List, b: List, key: callable) -> List:
    """
    Returns a list of all items, that are new in b when compared to a,
    using the key function to determine a unique identifier for list items
    :param a: "old" list
    :param b: "new" list
    :param key: function to map list items to a unique identifier
    :return: new list items
    """
    a_ids = set(map(key, a))
    b_ids = set(map(key, b))
    new_ids = b_ids - a_ids

    result = []
    for id in new_ids:
        item_in_b = list(filter(lambda x: key(x) == id, b))[0]
        result.append(item_in_b)
    return result


def approval_to_str(data: dict) -> str:
    id = data["id"]
    identifier = data["identifier"]
    current_version = data["currentVersion"]
    new_version = data["newVersion"]
    votes_required = data["votesRequired"]
    votes_received = data["votesReceived"]
    deadline = iso8601.parse_date(data["deadline"])
    message = data["message"]

    now_utc = datetime.now().replace(microsecond=0).astimezone(tz=timezone.utc)
    deadline_diff = timedelta(seconds=(deadline.replace(microsecond=0) - now_utc).total_seconds())

    deadline_abs_str = deadline.strftime('%m/%d %H:%M:%S')
    deadline_remaining_str = deadline_diff_to_str(deadline_diff)

    text = "\n".join([
        f"<b>{message}</b>",
        f"Id: {id}",
        f"Identifier: {identifier}",
        f"Version: {current_version} -> {new_version}",
        f"Votes: {votes_received}/{votes_required}",
        f"Expires: {deadline_abs_str} ({deadline_remaining_str})"
    ])

    return text


def deadline_diff_to_str(deadline_diff) -> str:
    units = []

    days = deadline_diff.days if deadline_diff.days else ""
    if days:
        units.append(f"{days}d")

    hours = deadline_diff.seconds // 3600
    if hours:
        units.append(f"{hours}h")

    minutes = (deadline_diff.seconds // 60) % 60
    if minutes:
        units.append(f"{minutes}m")

    seconds = int(deadline_diff.seconds) % 60
    if seconds:
        units.append(f"{seconds}s")

    return "".join(units)



import asyncio
import logging
import threading

LOGGER = logging.getLogger(__name__)


class RegularIntervalWorker:
    """
    Base class for a worker that executes a specific task in a regular interval.
    """

    def __init__(self, interval: float):
        self._interval = interval
        self._timer = None

    async def start(self):
        """
        Starts the worker
        """
        if self._timer is None:
            LOGGER.debug(f"Starting worker: {self.__class__.__name__}")
            self._schedule_next_run(0)

            # wait forever
            return await asyncio.Event().wait()
        else:
            LOGGER.debug("Already running, ignoring start() call")

    def stop(self):
        """
        Stops the worker
        """
        if self._timer is not None:
            self._timer.cancel()
        self._timer = None

    def _schedule_next_run(self, interval: int = None):
        """
        Schedules the next run
        """
        if self._timer is not None:
            self._timer.cancel()
        interval = interval if interval is not None else self._interval
        self._timer = threading.Timer(interval, self._worker_job)
        self._timer.start()

    def _worker_job(self):
        """
        The regularly executed task. Override this method.
        """
        try:
            loop = asyncio.new_event_loop()
            loop.run_until_complete(self._run())
        except Exception as e:
            LOGGER.error(e, exc_info=True)
        finally:
            self._schedule_next_run()

    async def _run(self):
        """
        The regularly executed task. Override this method.
        """
        raise NotImplementedError()



from prometheus_client import Summary, Counter
from prometheus_client.metrics import MetricWrapperBase

from keel_telegram_bot.const import *

COMMAND_TIME = Summary('command_processing_seconds', 'Time spent in a command handler', ['command'])
COMMAND_TIME_START = COMMAND_TIME.labels(command=COMMAND_START)
COMMAND_TIME_LIST_APPROVALS = COMMAND_TIME.labels(command=COMMAND_LIST_APPROVALS[0])
COMMAND_TIME_APPROVE = COMMAND_TIME.labels(command=COMMAND_APPROVE[0])
COMMAND_TIME_REJECT = COMMAND_TIME.labels(command=COMMAND_REJECT[0])
COMMAND_TIME_DELETE = COMMAND_TIME.labels(command=COMMAND_DELETE[0])

WATCHER_TIME = Summary('watcher_processing_seconds', 'Time spent in a watcher', ['type'])
APPROVAL_WATCHER_TIME = WATCHER_TIME.labels(type="approval")

NEW_PENDING_APPROVAL_COUNTER = Counter('keel_new_pending_approval',
                                       'Counts new pending approvals recognized by this bot')

REST_TIME = Summary('rest_endpoint_processing_seconds', 'Time spent in a rest command handler', ['endpoint'])
REST_TIME_WEBHOOK = REST_TIME.labels(endpoint=ENDPOINT_WEBHOOK)

KEEL_NOTIFICATION_COUNTER = Counter('keel_notifications', 'Counts notifications received from keel')
KEEL_APPROVAL_ACTION_COUNTER = Counter('keel_approval_action_counter',
                                       'Counts approval notificaion actions', ['action', 'identifier'])


def get_metrics() -> []:
    entries = set()
    for name, obj in globals().items():
        if isinstance(obj, MetricWrapperBase):
            entries.add(obj)

    return list(entries)


def format_metrics() -> str:
    def format_sample(sample):
        result = "  "
        if len(sample[0]) > 0:
            result += str(sample[0])
        if len(sample[1]) > 0:
            result += str(sample[1])

        if len(result) > 0:
            result += " "
        result += str(sample[2])

        return result

    def format_samples(samples):
        return "\n".join(list(map(format_sample, samples)))

    def format_metric(metric):
        name = metric._name
        samples = list(metric._samples())
        samples_text = format_samples(samples)

        return "{}:\n{}".format(name, samples_text)

    return "\n\n".join(map(format_metric, get_metrics()))



from keel_telegram_bot.util import approval_to_str
from tests import TestBase


class DummyTest(TestBase):

    def test_dummy(self):
        self.assertTrue(True)

    def test_approval_format(self):
        approval = {
            "id": "48d6da3e-e4c9-4d12-8562-b7975e805d80",
            "identifier": "deployment/local-path-storage/local-path-provisioner:v0.0.19",
            "currentVersion": "v0.0.17",
            "newVersion": "v0.0.19",
            "votesRequired": 1,
            "votesReceived": 0,
            "deadline": "2020-12-18 23:16:14.811933+00:00",
            "message": "New image is available for resource local-path-storage/local-path-provisioner (v0.0.17 -> v0.0.19).",
        }

        result = approval_to_str(approval)
        print(result)



import logging
import re

import voluptuous
from container_app_conf import ConfigBase
from container_app_conf.entry.bool import BoolConfigEntry
from container_app_conf.entry.dict import DictConfigEntry
from container_app_conf.entry.int import IntConfigEntry
from container_app_conf.entry.list import ListConfigEntry
from container_app_conf.entry.string import StringConfigEntry
from container_app_conf.entry.timedelta import TimeDeltaConfigEntry
from container_app_conf.source.env_source import EnvSource
from container_app_conf.source.toml_source import TomlSource
from container_app_conf.source.yaml_source import YamlSource
from voluptuous import Schema

NODE_MAIN = "keel-telegram-bot"

NODE_TELEGRAM = "telegram"

NODE_KEEL = "keel"
NODE_HOST = "host"
NODE_WEBHOOK = "webhook"

NODE_FILTERS = "filters"

NODE_STATS = "stats"
NODE_ENABLED = "enabled"
NODE_PORT = "port"


class Config(ConfigBase):

    def __new__(cls, *args, **kwargs):
        yaml_source = YamlSource(NODE_MAIN)
        toml_source = TomlSource(NODE_MAIN)
        data_sources = [
            EnvSource(),
            yaml_source,
            toml_source
        ]
        return super(Config, cls).__new__(cls, data_sources=data_sources)

    LOG_LEVEL = StringConfigEntry(
        description="Log level",
        key_path=[
            NODE_MAIN,
            "log_level"
        ],
        regex=re.compile(f"{'|'.join(logging._nameToLevel.keys())}", flags=re.IGNORECASE),
        default="WARNING",
    )

    TELEGRAM_BOT_TOKEN = StringConfigEntry(
        description="The telegram bot token to use",
        key_path=[
            NODE_MAIN,
            NODE_TELEGRAM,
            "bot_token"
        ],
        example="123456:ABC-DEF1234ghIkl-zyx57W2v1u123ew11",
        secret=True,
        required=True
    )

    TELEGRAM_ADMIN_USERNAMES = ListConfigEntry(
        item_type=StringConfigEntry,
        key_path=[
            NODE_MAIN,
            NODE_TELEGRAM,
            "admin_usernames"
        ],
        required=True,
        example=[
            "myadminuser",
            "myotheradminuser"
        ]
    )

    TELEGRAM_CHAT_IDS = ListConfigEntry(
        item_type=StringConfigEntry,
        key_path=[
            NODE_MAIN,
            NODE_TELEGRAM,
            "chat_ids"
        ],
        required=True,
        example=[
            12345678,
            87654321
        ]
    )

    KEEL_HOST = StringConfigEntry(
        description="Hostname of the keel HTTP endpoint",
        key_path=[
            NODE_MAIN,
            NODE_KEEL,
            NODE_HOST
        ],
        default="localhost",
        required=True
    )

    KEEL_PORT = IntConfigEntry(
        description="Port of the keel HTTP endpoint",
        key_path=[
            NODE_MAIN,
            NODE_KEEL,
            NODE_PORT
        ],
        default=9300,
        required=True
    )

    KEEL_SSL = BoolConfigEntry(
        description="Whether to use HTTPS or not",
        key_path=[
            NODE_MAIN,
            NODE_KEEL,
            "ssl"
        ],
        default=True
    )

    KEEL_USER = StringConfigEntry(
        description="Keel basic auth username",
        key_path=[
            NODE_MAIN,
            NODE_KEEL,
            "username"
        ],
        required=True
    )

    KEEL_PASSWORD = StringConfigEntry(
        description="Keel basic auth password",
        key_path=[
            NODE_MAIN,
            NODE_KEEL,
            "password"
        ],
        required=True,
        secret=True
    )

    MONITOR_INTERVAL = TimeDeltaConfigEntry(
        description="Interval to check for new pending approvals",
        key_path=[
            NODE_MAIN,
            "monitor",
            "interval"
        ],
        default="1m",
        required=True,
    )

    TELEGRAM_FILTERS = ListConfigEntry(
        description="Per chat-id filter to apply to the list of approvals",
        key_path=[
            NODE_MAIN,
            NODE_TELEGRAM,
            NODE_FILTERS,
        ],
        item_type=DictConfigEntry,
        item_args={
            "schema": Schema({
                voluptuous.Required('chat_id'): str,
                voluptuous.Required('identifier'): str,
            })
        },
        default=None,
    )

    STATS_ENABLED = BoolConfigEntry(
        description="Whether to enable prometheus statistics or not.",
        key_path=[
            NODE_MAIN,
            NODE_STATS,
            NODE_ENABLED
        ],
        default=True
    )

    STATS_PORT = IntConfigEntry(
        description="The port to expose statistics on.",
        key_path=[
            NODE_MAIN,
            NODE_STATS,
            NODE_PORT
        ],
        default=8000
    )



import asyncio
import json
import logging

import aiohttp
from aiohttp import web
from aiohttp.web_response import Response

from keel_telegram_bot.bot import KeelTelegramBot
from keel_telegram_bot.config import Config
from keel_telegram_bot.const import ENDPOINT_WEBHOOK

LOGGER = logging.getLogger(__name__)
routes = web.RouteTableDef()


class WebsocketServer:
    bot = None

    def __init__(self, config: Config, bot: KeelTelegramBot):
        self.config = config
        WebsocketServer.bot = bot

    async def start(self):
        host = "0.0.0.0"  # self.config.SERVER_HOST.value,
        port = 5000  # self.config.SERVER_PORT.value
        LOGGER.info(f"Starting webserver on {host}:{port} ...")

        app = self._create_app()
        runner = aiohttp.web.AppRunner(app)
        await runner.setup()
        site = aiohttp.web.TCPSite(
            runner,
            host=host,
            port=port,
        )
        await site.start()

        # wait forever
        return await asyncio.Event().wait()

    def _create_app(self) -> web.Application:
        app = web.Application(middlewares=[])
        app.add_routes(routes)
        return app

    # @app.route('/', methods=['GET', 'POST'])
    # @app.route('/<path:path>', methods=['GET', 'POST'])
    # @routes.get(ENDPOINT_WEBHOOK)
    @routes.post(ENDPOINT_WEBHOOK)
    # @time(REST_TIME_WEBHOOK)
    # @staticmethod
    async def catch_all(self) -> Response:
        if not self.can_read_body:
            LOGGER.error("Request has no body!")
            return Response(text="Failed to read request")
        body = await self.content.read()
        if len(body) <= 0:
            LOGGER.error("Request has no body!")
            return Response(text="Request body is empty!")

        data = json.loads(body)
        await WebsocketServer.bot.on_notification(data)
        return Response(text="OK")



import logging
from typing import List, Any

from telegram import Update, ReplyKeyboardRemove, ReplyKeyboardMarkup, KeyboardButton
from telegram.ext import CallbackContext

from keel_telegram_bot.const import CANCEL_KEYBOARD_COMMAND
from keel_telegram_bot.util import send_message, fuzzy_match

LOGGER = logging.getLogger(__name__)


class ReplyKeyboardHandler:
    # this map is used to remember from which users we
    # are currently awaiting a response message
    awaiting_response = {}

    def __init__(self):
        pass

    async def on_message(self, update: Update, context: CallbackContext):
        user_id = update.effective_user.id
        text = update.effective_message.text

        if user_id not in self.awaiting_response:
            return

        data = self.awaiting_response[user_id]
        if text not in data["valid_responses"]:
            return

        LOGGER.debug("Awaited response from user {} received: {}".format(user_id, text))
        try:
            await data["callback"](update, context, text, data["callback_data"])
            self.awaiting_response.pop(user_id)
        except Exception as e:
            LOGGER.exception(e)

    async def _on_user_selection(self, update: Update, context: CallbackContext, message: str, data: dict):
        """
        Called when a user selection is awaited and user message arrives
        :param update:
        :param context:
        :param message:
        :param data:
        """
        await self.await_user_selection(update, context, message, data["choices"], data["key"],
                                        data["callback"], data["callback_data"])

    async def await_user_selection(self, update: Update, context: CallbackContext,
                                   selection: str or None, choices: List[Any], key: callable,
                                   callback: callable, callback_data: dict = None):
        """
        Sends a ReplyKeyboard to the user and waits for a valid selection.
        :param update: Update
        :param context: CallbackContext
        :param selection: the current user selection (if any)
        :param choices: list of choices to select from
        :param key: function to create unique string key for a choice
        :param callback: the function to call, when a selection was made
        :param callback_data: data to pass to the callback function
        """
        bot = context.bot
        chat_id = update.effective_chat.id
        message_id = update.effective_message.message_id
        user_id = update.effective_user.id

        fuzzy_matches = fuzzy_match(selection, choices=choices, key=key, limit=5)

        # check if something matches perfectly
        perfect_matches = list(filter(lambda x: x[1] == 100, fuzzy_matches))
        if len(perfect_matches) == 1:
            choice = perfect_matches[0][0]
            await callback(update, context, choice, callback_data)
            return

        # send reply keyboard with fuzzy matches to user
        keyboard_texts = list(map(lambda x: "{}".format(key(x[0])), fuzzy_matches))
        keyboard = self.build_reply_keyboard(keyboard_texts)
        text = "No unique perfect match found, please select one of the menu options"
        self.await_response(
            user_id=user_id,
            options=keyboard_texts,
            callback=self._on_user_selection,
            callback_data={
                "choices": choices,
                "key": key,
                "callback": callback,
                "callback_data": callback_data,
            })
        await send_message(bot, chat_id, text, parse_mode="MARKDOWN", reply_to=message_id, menu=keyboard)

    def await_response(self, user_id: str, options: List[str], callback_data: dict, callback):
        """
        Remember, that we are awaiting a response message from a user
        :param user_id: the user id
        :param options: a list of messages that the user can send as a valid response
        :param callback: function to call with callback data
        :param callback_data: data to pass to callback
        """
        if user_id in self.awaiting_response:
            raise AssertionError("Already awaiting response to a previous query from user {}".format(user_id))

        self.awaiting_response[user_id] = {
            "valid_responses": options,
            "callback": callback,
            "callback_data": callback_data
        }

    async def cancel_keyboard_callback(self, update: Update, context: CallbackContext):
        bot = context.bot
        chat_id = update.effective_chat.id
        user_id = update.effective_user.id
        message_id = update.effective_message.message_id

        text = "Cancelled"
        await send_message(bot, chat_id, text, reply_to=message_id, menu=ReplyKeyboardRemove(selective=True))

        if user_id in self.awaiting_response:
            self.awaiting_response.pop(user_id)

    @staticmethod
    def build_reply_keyboard(items: List[str]) -> ReplyKeyboardMarkup:
        """
        Builds a menu to select an item from a list
        :param items: list of items to choose from
        :return: reply markup
        """
        items.append(CANCEL_KEYBOARD_COMMAND)
        keyboard = list(map(lambda x: KeyboardButton(x), items))
        # NOTE: the "selective=True" requires the menu to be sent as a reply
        # (or with an @username mention)
        return ReplyKeyboardMarkup.from_column(keyboard, one_time_keyboard=True, selective=True)



import logging

from keel_telegram_bot.api_client import KeelApiClient
from keel_telegram_bot.bot import KeelTelegramBot
from keel_telegram_bot.config import Config
from keel_telegram_bot.monitoring import RegularIntervalWorker
from keel_telegram_bot.stats import APPROVAL_WATCHER_TIME, NEW_PENDING_APPROVAL_COUNTER
from keel_telegram_bot.util import filter_new_by_key

LOGGER = logging.getLogger(__name__)


class Monitor(RegularIntervalWorker):

    def __init__(self, config: Config, api_client: KeelApiClient, bot: KeelTelegramBot):
        interval_seconds = config.MONITOR_INTERVAL.value.total_seconds()
        super().__init__(interval_seconds)
        self._config = config
        self._api_client = api_client
        self._bot = bot
        self._old = None

    @APPROVAL_WATCHER_TIME.time()
    async def _run(self):
        """
        Called repeatedly
        """
        active = self._api_client.get_approvals(rejected=False, archived=False)

        try:
            # update existing messages
            await self._bot.update_messages()
        except Exception as ex:
            LOGGER.exception(ex)

        if self._old is None:
            self._old = active
            return

        new_pending = filter_new_by_key(self._old, active, key=lambda x: x["id"])
        self._old = active

        for item in new_pending:
            NEW_PENDING_APPROVAL_COUNTER.inc()
            await self._bot.on_new_pending_approval(item)



import asyncio
import logging
import os
import sys

from container_app_conf.formatter.toml import TomlFormatter
from prometheus_client import start_http_server

from keel_telegram_bot.api_client import KeelApiClient
from keel_telegram_bot.bot import KeelTelegramBot
from keel_telegram_bot.config import Config
from keel_telegram_bot.monitoring.monitor import Monitor
from keel_telegram_bot.webserver import WebsocketServer

parent_dir = os.path.abspath(os.path.join(os.path.abspath(__file__), "..", ".."))
sys.path.append(parent_dir)

logging.basicConfig(format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')
LOGGER = logging.getLogger(__name__)


def main():
    config = Config()

    log_level = logging._nameToLevel.get(str(config.LOG_LEVEL.value).upper(), config.LOG_LEVEL.default)
    logging.getLogger("keel_telegram_bot").setLevel(log_level)

    LOGGER.debug("Config:\n{}".format(config.print(TomlFormatter())))

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    # start prometheus server
    if config.STATS_ENABLED.value:
        start_http_server(config.STATS_PORT.value)

    api_client = KeelApiClient(
        config.KEEL_HOST.value,
        config.KEEL_PORT.value,
        config.KEEL_SSL.value,
        config.KEEL_USER.value,
        config.KEEL_PASSWORD.value,
    )

    bot = KeelTelegramBot(config, api_client)
    monitor = Monitor(config, api_client, bot)
    server = WebsocketServer(config, bot)

    tasks = asyncio.gather(
        bot.start(),
        monitor.start(),
        server.start(),
    )

    loop.run_until_complete(tasks)
    loop.run_forever()

    # monitor.stop()
    # bot.stop()


if __name__ == '__main__':
    main()



import os
import subprocess

from setuptools import setup, find_packages

from keel_telegram_bot import __version__

GIT_BRANCH = subprocess.check_output(["git", "rev-parse", "--abbrev-ref", "HEAD"])
GIT_BRANCH = GIT_BRANCH.decode()  # convert to standard string
GIT_BRANCH = GIT_BRANCH.rstrip()  # remove unnecessary whitespace

if GIT_BRANCH == "master":
    DEVELOPMENT_STATUS = "Development Status :: 5 - Production/Stable"
    VERSION_NAME = __version__
elif GIT_BRANCH == "beta":
    DEVELOPMENT_STATUS = "Development Status :: 4 - Beta"
    VERSION_NAME = "%s-beta" % __version__
elif GIT_BRANCH == "dev":
    DEVELOPMENT_STATUS = "Development Status :: 3 - Alpha"
    VERSION_NAME = "%s-dev" % __version__
elif (os.environ.get("TRAVIS_BRANCH", None) == os.environ.get("TRAVIS_TAG", None) == "v{}".format(__version__)) or \
        "GITHUB_RELEASE" in os.environ:
    # travis tagged release branch
    DEVELOPMENT_STATUS = "Development Status :: 5 - Production/Stable"
    VERSION_NAME = __version__
else:
    print("Unknown git branch, using pre-alpha as default")
    DEVELOPMENT_STATUS = "Development Status :: 2 - Pre-Alpha"
    VERSION_NAME = "%s-%s" % (__version__, GIT_BRANCH)


def readme_type() -> str:
    import os
    if os.path.exists("README.rst"):
        return "text/x-rst"
    if os.path.exists("README.md"):
        return "text/markdown"


def readme() -> [str]:
    with open('README.md') as f:
        return f.read()


def locked_requirements(section):
    """
    Look through the 'Pipfile.lock' to fetch requirements by section.
    """
    import json

    with open('Pipfile.lock') as pip_file:
        pipfile_json = json.load(pip_file)

    if section not in pipfile_json:
        print("{0} section missing from Pipfile.lock".format(section))
        return []

    return [package + detail.get('version', "")
            for package, detail in pipfile_json[section].items()]


setup(
    name='keel-telegram-bot',
    version=VERSION_NAME,
    description='A telegram bot for keel.sh',
    long_description=readme(),
    long_description_content_type=readme_type(),
    license='AGPLv3+',
    author='Markus Ressel',
    author_email='mail@markusressel.de',
    url='https://github.com/markusressel/keel-telegram-bot',
    packages=find_packages(),
    # python_requires='>=3.4',
    classifiers=[
        DEVELOPMENT_STATUS,
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.8'
    ],
    install_requires=locked_requirements('default'),
    tests_require=locked_requirements('develop'),
    entry_points={
        'console_scripts': [
            'keel-telegram-bot = keel_telegram_bot.main:main'
        ]
    }
)



from telegram import Update
from telegram.ext import CallbackContext
from telegram_click.permission.base import Permission

from keel_telegram_bot.config import Config


class _ConfigAdmins(Permission):

    def __init__(self):
        self._config = Config()

    async def evaluate(self, update: Update, context: CallbackContext) -> bool:
        from_user = update.effective_message.from_user
        return from_user.username in self._config.TELEGRAM_ADMIN_USERNAMES.value


CONFIG_ADMINS = _ConfigAdmins()


class _ConfigChatId(Permission):

    def __init__(self):
        self._config = Config()

    async def evaluate(self, update: Update, context: CallbackContext) -> bool:
        chat_id = update.effective_message.chat_id
        return str(chat_id) in self._config.TELEGRAM_CHAT_IDS.value


CONFIGURED_CHAT_ID = _ConfigChatId()



from keel_telegram_bot import util
from tests import TestBase


class BotTest(TestBase):

    def test_is_filtered_for_real(self):
        chat_id = "123456789"
        identifier = "deployment/local-path-storage/local-path-provisioner:v0.0.22"

        filters = [
            {
                "chat_id": "123456789",
                "identifier": ".*satisfactory.*"
            },
            {
                "chat_id": "-123456789123456",
                "identifier": ".*"
            }
        ]

        result = util._is_filtered_for(filters, chat_id, identifier)

        self.assertTrue(result)

    def test_is_filtered_for_real2(self):
        chat_id = "-123456789123456"
        identifier = "deployment/local-path-storage/local-path-provisioner:v0.0.22"

        filters = [
            {
                "chat_id": "123456789",
                "identifier": ".*"
            },
            {
                "chat_id": "-123456789123456",
                "identifier": ".*satisfactory.*"
            }
        ]

        result = util._is_filtered_for(filters, chat_id, identifier)

        self.assertTrue(result)

    def test_is_filtered_for_false(self):
        chat_id = "123456"
        identifier = "deployment/gameservers/satisfactory"

        filters = [
            {
                "chat_id": "123456",
                "identifier": ".*satisfactory.*"
            }
        ]

        result = util._is_filtered_for(filters, chat_id, identifier)

        self.assertFalse(result)

    def test_is_filtered_for_true(self):
        chat_id = "123456"
        identifier = "deployment/gameservers/satisfactory"

        filters = [
            {
                "chat_id": "123456",
                "identifier": ".*wiki.*"
            }
        ]

        result = util._is_filtered_for(filters, chat_id, identifier)

        self.assertTrue(result)

    def test_is_filtered_for_no_match(self):
        chat_id = "123456"
        identifier = "deployment/gameservers/satisfactory"

        filters = [
            {
                "chat_id": "654321",
                "identifier": ".*satisfactory.*"
            }
        ]

        result = util._is_filtered_for(filters, chat_id, identifier)

        self.assertFalse(result)
